/*
** File: admxrc3_mmap.c  
** Project: ADMXRC3 API module library
** Purpose: Implements Linux-specific functions for mapping memory in
**          ADMXRC3 API module library.
**
** (C) Copyright Alpha Data 2012
*/

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>
#include <pthread.h>

#include "admxrc3_platform.h"
#include <ioctl_dflinux.h>

#define TRACE 0

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

struct _MappedRegion;
typedef struct _MappedRegion MappedRegion;

struct _MappedRegion {
  MappedRegion* pPrev;
  MappedRegion* pNext;
  ADMXRC3_HANDLE hDevice;
  void* pAlignedVirtual;
  void* pBaseVirtual;
  size_t alignedLength;
};

static struct _MappedList {
  MappedRegion* pHead;
  MappedRegion* pTail;
} g_mappedList;
static pthread_mutex_t g_mappedListLock = PTHREAD_MUTEX_INITIALIZER;
static int g_bInitialized = FALSE;

static void
initialize(
  void)
{
  pthread_mutex_lock(&g_mappedListLock);
  if (!g_bInitialized) {
#if TRACE
    fprintf(stderr, "initialize: initializing globals\n");
#endif
    g_mappedList.pHead = NULL;
    g_mappedList.pTail = NULL;
    g_bInitialized = TRUE;
  }
  pthread_mutex_unlock(&g_mappedListLock);
}

static void
addRegion(
  MappedRegion* pMappedRegion)
{
  pthread_mutex_lock(&g_mappedListLock);
  pMappedRegion->pPrev = g_mappedList.pTail;
  pMappedRegion->pNext = NULL;
  if (NULL == g_mappedList.pHead) {
    g_mappedList.pHead = pMappedRegion;
    g_mappedList.pTail = pMappedRegion;
  } else {
    g_mappedList.pTail->pNext = pMappedRegion;
    g_mappedList.pTail = pMappedRegion;
  }
  pthread_mutex_unlock(&g_mappedListLock);
}

/* Must be called while holding the semaphore */
static void
removeRegion(
  MappedRegion* pMappedRegion)
{
  if (NULL == pMappedRegion->pPrev) {
    g_mappedList.pHead = pMappedRegion->pNext;
  } else {
    pMappedRegion->pPrev->pNext = pMappedRegion->pNext;
  }
  if (NULL == pMappedRegion->pNext) {
    g_mappedList.pTail = pMappedRegion->pPrev;
  } else {
    pMappedRegion->pNext->pPrev = pMappedRegion->pPrev;
  }
  pMappedRegion->pPrev = NULL;
  pMappedRegion->pNext = NULL;
}

static MappedRegion*
findAndRemoveRegion(
  ADMXRC3_HANDLE hDevice,
  const void* pBaseVirtual,
  bool_t* pbFoundHandle)
{
  MappedRegion* p;
  bool_t bFoundHandle = FALSE;

  pthread_mutex_lock(&g_mappedListLock);
  p = g_mappedList.pHead;
  while (NULL != p) {
    if (p->hDevice == hDevice) {
      bFoundHandle = TRUE;
      if (p->pBaseVirtual == pBaseVirtual) {
        break;
      }
    }
    p = p->pNext;
  }
  if (NULL != p) {
#if TRACE
    fprintf(stderr, "findAndRemoveRegion: removing region, hDevice=%d pBaseVirtual=%p pRegion=%p\n", hDevice, pBaseVirtual, p);
#endif
    removeRegion(p);
  } else {
#if TRACE
    fprintf(stderr, "findAndRemoveRegion: failed to find region, hDevice=%d pBaseVirtual=%p pRegion=%p\n", hDevice, pBaseVirtual, p);
#endif
  }
  pthread_mutex_unlock(&g_mappedListLock);

  *pbFoundHandle = bFoundHandle;
  return p;
}

ADMXRC3_STATUS
admxrc3MapRegion(
  ADMXRC3_HANDLE hAdmxrc3,
  unsigned int regionTag,
  uint64_t offset,
  uint64_t length,
  void** ppVirtualBase)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  DfLinuxIoctlGetMmapAddress io;
  uint8_t* p = NULL;
  int err, ioCode;
  long pageSize;
  uint64_t address;
  off_t pageMask, pageStart;
  size_t pageLength, pageOffset;
  MappedRegion* pMappedRegion = NULL;

  initialize();

  assert(sizeof(off_t) == sizeof(unsigned long));

  /* Sanity checks */
  if (0 == length || offset + length < offset) {
    return ADMXRC3_INVALID_REGION;
  }

  pMappedRegion = (MappedRegion*)malloc(sizeof(MappedRegion));
  if (NULL == pMappedRegion) {
    return ADMXRC3_NO_MEMORY;
  }

  ioCode = _IOC(_IOC_READ | _IOC_WRITE, 150, DFLINUX_IOCTLCODE_GETMMAPADDRESS, 0);
  io.in.regionTag = regionTag;
  err = ioctl(hAdmxrc3, ioCode, &io);
  if (0 != err) {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        status = ADMXRC3_INVALID_HANDLE;
        break;

      default:
        fprintf(stderr, "admxrc3MapRegion: ioctl() failed, errno=%d (%s)\n", errno, strerror(errno));
        status = ADMXRC3_UNEXPECTED_ERROR;
        break;
      }
    } else {
      switch ((DfLinuxMmapStatus)err) {
      case DfLinuxMmapInvalidRegion:
        status = ADMXRC3_INVALID_REGION;
        break;

      case DfLinuxMmapInvalidTag:
        status = ADMXRC3_INVALID_INDEX;
        break;

      case DfLinuxMmapNoMemory:
        status = ADMXRC3_NO_MEMORY;
        break;

      case DfLinuxMmapRegionTooLarge:
        status = ADMXRC3_REGION_TOO_LARGE;
        break;

      case DfLinuxMmapAccessDenied:
        status = ADMXRC3_ACCESS_DENIED;
        break;

      case DfLinuxMmapGeneralFailure:
      default:
        status = ADMXRC3_UNEXPECTED_ERROR;
        break;
      }
    }
  }

  if (ADMXRC3_SUCCESS != status) {
    goto out;
  }

  pageSize = sysconf(_SC_PAGESIZE);
  if (pageSize == (unsigned long)-1) {
    fprintf(stderr, "admxrc3MapRegion: get page size from sysconf() failed, errno=%d (%s)\n", errno, strerror(errno));
    status = ADMXRC3_UNEXPECTED_ERROR;
    goto out;
  }
  pageMask = (off_t)(pageSize - 1U);
  address = io.out.address + offset;
  if (length + offset > io.out.length) {
    status = ADMXRC3_INVALID_REGION;
    goto out;
  }
  pageStart = (off_t)address & ~pageMask;
  pageOffset = (size_t)address & (size_t)pageMask;
  pageLength = (size_t)((size_t)length + pageOffset + (size_t)pageMask) & ~(size_t)pageMask;
#if TRACE
  fprintf(stderr, "admxrc3MapRegion: calling mmap, hDevice=%d length=%p address=0x%llx\n",
    hAdmxrc3, (void*)pageLength, (unsigned long long)pageStart);
#endif
  p = (uint8_t*)mmap(NULL, pageLength, PROT_READ | PROT_WRITE, MAP_SHARED, hAdmxrc3, pageStart);
  if (MAP_FAILED == p) {
    switch (errno) {
    default:
      fprintf(stderr, "admxrc3MapRegion: mmap() failed, errno=%d (%s)\n", errno, strerror(errno));
      status = ADMXRC3_UNEXPECTED_ERROR;
      goto out;
    }
  }
  pMappedRegion->hDevice = hAdmxrc3;
  pMappedRegion->alignedLength = pageLength;
  pMappedRegion->pAlignedVirtual  = p;
  p += pageOffset;
  pMappedRegion->pBaseVirtual = p;
#if TRACE
  fprintf(stderr, "admxrc3MapRegion: adding region, hDevice=%d pRegion%p\n", hAdmxrc3, pMappedRegion);
#endif
  addRegion(pMappedRegion);

out:
  if (ADMXRC3_SUCCESS == status) {
    *ppVirtualBase = (void*)p;
  } else {
    if (NULL != pMappedRegion) {
      free(pMappedRegion);
    }
  }
  return status;
}

ADMXRC3_STATUS
admxrc3UnmapRegion(
  ADMXRC3_HANDLE hAdmxrc3,
  void* pVirtualBase)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  MappedRegion* p;
  bool_t bFoundHandle = FALSE;

  initialize();

  p = findAndRemoveRegion(hAdmxrc3, pVirtualBase, &bFoundHandle);
  if (NULL == p) {
    return bFoundHandle ? ADMXRC3_INVALID_REGION : ADMXRC3_INVALID_HANDLE;
  }

#if TRACE
  fprintf(stderr, "admxrc3UnmapRegion: calling munmap, aligned=%p length=%p\n",
    p->pAlignedVirtual, (void*)p->alignedLength);
#endif
  if (0 != munmap(p->pAlignedVirtual, p->alignedLength)) {
    switch (errno) {
    case EINVAL:
      status = ADMXRC3_INVALID_REGION;
      break;

    default:
      fprintf(stderr, "admxrc3UnmapRegion: munmap() failed, base=%p length=%p errno=%d (%s)\n",
        p->pAlignedVirtual, (void*)p->alignedLength, errno, strerror(errno));
      status = ADMXRC3_UNEXPECTED_ERROR;
      break;
    }
  }
  free(p);
  
  return status;
}

void
admxrc3UnmapAllRegions(
  ADMXRC3_HANDLE hAdmxrc3)
{
  MappedRegion* p, * pNext;

  initialize();

  pthread_mutex_lock(&g_mappedListLock);
  p = g_mappedList.pHead;
  while (NULL != p) {
    pNext = p->pNext;
    if (p->hDevice == hAdmxrc3) {
#if TRACE
      fprintf(stderr, "admxrc3UnmapAllRegions: removing region, hDevice=%d pRegion=%p\n", hAdmxrc3, p);
#endif
      removeRegion(p);
#if TRACE
      fprintf(stderr, "admxrc3UnmapAllRegions: calling munmap, aligned=%p length=%p\n",
        p->pAlignedVirtual, (void*)p->alignedLength);
#endif
      if (0 != munmap(p->pAlignedVirtual, p->alignedLength)) {
        fprintf(stderr, "admxrc3UnmapAllRegions: munmap() failed, base=%p length=%p errno=%d (%s)\n",
          p->pAlignedVirtual, (void*)p->alignedLength, errno, strerror(errno));
      }
      free(p);
    }
    p = pNext;
  }
  pthread_mutex_unlock(&g_mappedListLock);
}
